#!/usr/bin/python3
# -*- coding: utf-8 -*-

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.lines import Line2D

from crystal import Crystal
from crystal_funcs import triplet_calculation

__author__ = "Rafaela Felix"
__credits__ = {"Rafaela Felix", "Sergio Morelhão"}
__version__ = "1.0"
__maintainer__ = "Rafaela Felix"
__email__ = "rafaelafelixp@usp.br"
__status__ = "Production"


class AMD:

    """**Construction of asymmetry diagrams (AMDs) and compatibility analysis.**"""

    def __init__(self, fexp: str, mnames: list = []):

        """Given the experimental asymmetries and structural models, returns a new `amd` object.

        Args:
            fexp (str): File name of experimental asymmetries.
            mnames (list): List of .in file names for structural models.

        Notes:
            fexp file presents hkl, asymmetry type and diffraction geometry (in this order) - `.red` generate by
            ExpData.save_output(). Should present the primary indexes in 1st line (skipping comments) and beam energy.
        """

        self.matrix = None
        self._mult = None
        self.hkl = np.loadtxt(fexp, usecols=(0, 1, 2), skiprows=2)
        self.asy = np.loadtxt(fexp, usecols=3, dtype=str, skiprows=2)
        self.s = np.loadtxt(fexp, usecols=4, skiprows=2)
        self.primary = np.loadtxt(fexp, usecols=(0, 1, 2), max_rows=1)
        self.energy = np.loadtxt(fexp, usecols=3, max_rows=1)
        self.names = mnames

        self.HKL = [str(int(self.hkl[i, 0])) + ' ' + str(int(self.hkl[i, 1])) + ' ' + str(int(self.hkl[i, 2])) for i in
                    range(len(self.hkl))]
        self.numeric_asy = [1 if t == 'HL' else 0 for t in self.asy]
        self._n = int(np.ceil(np.sqrt(len(self.asy))))

        label = np.append(self.HKL, np.zeros(self._n ** 2 - len(self.hkl), dtype=str)).reshape((self._n, self._n))[::-1]
        label = np.flip(label, axis=0)
        label[label == ''] = ' - '

        self.labels = label

    def exp_plot(self, hl: str or tuple = 'cyan', lh: str or tuple = 'gold'):

        """**Diagram of experimental asymmetries with respective indexation**.

        Args:
            hl (str or tuple): Color of HL (high/low) asymmetries in diagram. Default: 'cyan'.
            lh (str or tuple): Color of LH (low/high) asymmetries in diagram. Default: 'gold'.

        Notes:
            hl and lh are color names or RGB values (r, g, b) ranging from 0 to 1.
        """

        M = np.append(self.numeric_asy, np.zeros(self._n ** 2 - len(self.hkl))).reshape((self._n, self._n))

        cmap = plt.get_cmap('PRGn').copy()  # Colormap
        cmap.set_under(lh)
        cmap.set_over(hl)

        custom_lines = [Line2D([0], [0], marker='s', color=hl, lw=0.01, markersize=10),  # Customized legend
                        Line2D([0], [0], marker='s', color=lh, lw=0.01, markersize=10)]

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

        ax.pcolormesh(M, cmap=cmap, vmin=0.5, vmax=0.6)
        ax.hlines(y=np.arange(0, self._n + 1), xmin=0, xmax=self._n, color="k", linewidth=.5)
        ax.vlines(x=np.arange(0, self._n + 1), ymin=0, ymax=self._n, color="k", linewidth=.5)

        for y in range(0, self._n):
            for x in range(0, self._n):
                ax.text(x + 0.5, y + 0.5, str(self.labels[y, x]), horizontalalignment='center',
                        verticalalignment='center', weight='bold')

        ax.set_aspect('equal')
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.legend(custom_lines, ['HL', 'LH'], loc='upper left', bbox_to_anchor=(1, 1), prop={'size': 14})

        plt.show()

    def labels_plot(self):

        """**Diagram with MD-case positions.**"""

        cmap = mpl.colors.ListedColormap(['whitesmoke'])

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.pcolormesh(np.ones((self._n, self._n)), cmap=cmap)

        for y in range(0, self._n):
            for x in range(0, self._n):
                ax.text(x + 0.5, y + 0.5, str(self.labels[y, x]),
                        horizontalalignment='center', verticalalignment='center', weight='bold')

        ax.set_aspect('equal')
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

        ax.hlines(y=np.arange(0, self._n + 1), xmin=0, xmax=self._n, color="k", linewidth=.5)
        ax.vlines(x=np.arange(0, self._n + 1), ymin=0, ymax=self._n, color="k", linewidth=.5)

        plt.show()

    def matching_matrix(self):

        """**Generates matching matrix.**"""

        GH = self.primary - self.hkl  # coupling reflection

        i = 0
        for name in self.names:

            crystal = Crystal(name)
            FG = crystal.Fhkl(self.energy, self.primary)
            FH = crystal.Fhkl(self.energy, self.hkl)
            FGH = crystal.Fhkl(self.energy, GH)

            cosPsi = triplet_calculation(FG, FH, FGH)[0]

            if i > 0:
                cosT = np.row_stack((cosT, cosPsi))
            else:
                cosT = cosPsi[:]

            i += 1

        data = cosT*self.s
        data[data > 0] = 1
        data[data < 0] = 0

        m = int(np.sqrt(len(self.names)))
        M = np.zeros((m*self._n, m*self._n))

        for i in range(0, m):
            var_y = data[i*m:i*m+m, :]
            for j in range(0, m):
                var_x = var_y[j, :]
                R = np.zeros(self._n**2)
                for k in range(len(var_x)):
                    if var_x[k] == self.numeric_asy[k]:
                        R[k] = 1
                kp = len(var_x)
                for l in range(kp, kp + self._n ** 2 - len(self.asy)):
                    R[l] = 0.5

                R = np.reshape(R, (self._n, self._n))
                M[self._n * i:self._n * (i + 1), self._n * j:self._n * (j + 1)] = R[:, :]

        self._mult = self._n*m
        self.matrix = M

    def heatmap_plot(self, xmax: float, xmin:float, dx: float,
                     ymax: float, ymin: float, dy: float, xlabel: str = 'x', ylabel: str = 'y',
                     color_empty: str or tuple = 'whitesmoke', incp: str or tuple = 'red', cp: str or tuple = 'lime'):

        """**Plot asymmetry matching diagram.**

        Args:
            xmax (float): Maximum value in horizontal axis.
            xmin (float): Minimum value in horizontal axis.
            dx (float): Step in horizontal axis.
            ymax (float): Maximum value in vertical axis.
            ymin (float): Minimum value in vertical axis.
            dy (float): Step in vertical axis.
            xlabel (str): Label of horizontal axis. Default: 'x'.
            ylabel (str): Label of vertical axis. Default: 'y'.
            color_empty (str or tuple): Color of empty cells in matching matrix. Default: 'whitesmoke'.
            incp (str or tuple): Color of incompatible cases. Default: 'red'.
            cp (str or tuple): Color of compatible cases. Default: 'lime'.

        Notes:
            color_empty, incp and cp are color names or RGB values (r, g, b) ranging from 0 to 1.
        """

        # cmap = plt.get_cmap('PRGn').copy()  # Colormap
        cmap = mpl.colors.ListedColormap([incp, color_empty, cp])
        cmap.set_under(incp)
        cmap.set_over(cp)

        custom_lines = [Line2D([0], [0], marker='s', color=incp, lw=0.01, markersize=10),  # Customized legend
                        Line2D([0], [0], marker='s', color=cp, lw=0.01, markersize=10)]

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

        ax.pcolormesh(self.matrix, cmap=cmap, vmin=0.4, vmax=0.6)
        ax.hlines(y=np.arange(0, self._mult), xmin=0, xmax=self._mult, color="grey", linewidth=.5)
        ax.vlines(x=np.arange(0, self._mult), ymin=0, ymax=self._mult, color="grey", linewidth=.5)
        ax.hlines(y=np.arange(0, self._mult, self._n), xmin=0, xmax=self._mult, color="black", linewidth=1)
        ax.vlines(x=np.arange(0, self._mult, self._n), ymin=0, ymax=self._mult, color="black", linewidth=1)

        ax.set_xticks(ticks=np.arange(self._n/2, self._mult, self._n),
                      labels=np.round(np.arange(xmin, xmax + dx, dx), 1))
        ax.set_yticks(ticks=np.arange(self._n/2, self._mult, self._n),
                      labels=np.round(np.arange(ymin, ymax + dy, dy), 1))

        ax.set_aspect('equal')
        ax.legend(custom_lines, ['F', 'T'], loc='upper left', bbox_to_anchor=(1, 1), prop={'size': 14})

        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        plt.show()




