#!/usr/bin/python3

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly
import plotly.express as px
import plotly.subplots as sp

from funcs import delete_multiple
from scatter import *
from scipy.ndimage import gaussian_filter1d
from scipy.signal import find_peaks
from tqdm import tqdm

from crystal import Crystal
from expdata_funcs import dataframe_BClines, fwhm, peak_fit, bootstrap_resampling

__author__ = "Rafaela Felix"
__credits__ = {"Rafaela Felix", "Sergio Morelhão"}
__version__ = "1.0"
__maintainer__ = "Rafaela Felix"
__email__ = "rafaelafelixp@usp.br"
__status__ = "Production"


class ExpData:

    """**Analysis of Renninger scans.**"""

    def __init__(self, E: float, G: list, fname: str, colx: int = 0, coly: int = 1, name: str = ''):

        """Given the experimental data, returns a new `exp` object.

        Args:
            E (float): Beam energy (eV).
            G (list): Indices of primary reflection (eg, [1, 1, 1]).
            fname (str): Experimental data file name.
            colx (int): Column number with angle values on data file. 1st column, by default.
            coly (int): Column number with intensity values. 2nd column, by default.
            name: Label for analysis (e.g. 'pure', 'doped'). Default: None.
        """

        self.asy = None
        self.slope_data = None
        self.index = None
        self.BC = None
        self.region = None
        self.peaks = None
        self.phi = np.loadtxt(fname, usecols=colx)
        self.counts = np.loadtxt(fname, usecols=coly)
        self.energy = E
        self.primary = np.array(G)
        self.name = name

    def plot(self) -> plotly.graph_objects:

        """**Plot the experimental data.**

        Returns:
            plotly.graph_objects: Interactive plot.
        """

        df_exp = pd.DataFrame(np.column_stack((self.phi, self.counts)),
                              columns=['phi', 'I'])

        fig = px.scatter(df_exp, x='phi', y='I',
                         labels={'phi': '𝜙 (deg)', 'I': 'Counts'})
        fig.data[0].update(mode='markers+lines')
        fig.update_traces(marker=dict(color='black', size=6),
                          line=dict(color='black', width=2))

        fig.update_yaxes(title_text='Counts',
                         range=[0, 1.1 * self.counts.max()])
        fig.update_xaxes(range=[self.phi.min(), self.phi.max()])
        fig.update_layout(legend_title_text='', template='plotly_white',
                          font=dict(size=16))
        fig.show()

    def BC_plot(self, fname: str, M: list, Fmin: float) -> plotly.graph_objects:

        """**Plot the experimental data and BC lines.**

        Args:
            fname (str): Structure file name (the same used by Crystal class).
            M (list): Reference direction (eg, [1, 0, 0]).
            Fmin (float): Minimum cutoff value for W (absolute).

        Returns:
            plotly.graph_objects: Interactive plot.
        """

        M = np.array(M)
        struc = Crystal(fname)
        struc.klines(self.energy, self.primary, M, Fmin)

        fname = "G_" + "".join((self.primary.astype(int)).astype(str))
        fname += "_M_" + "".join((M.astype(int)).astype(str))
        fname += "_E_" + str(self.energy) + ".dat"
        thG, FHFGH, df_in = dataframe_BClines('IN_' + fname)
        thG, FHFGH, df_out = dataframe_BClines('OUT_' + fname)

        subfigure = sp.make_subplots(rows=2, cols=1,
                                     shared_xaxes=True,
                                     vertical_spacing=0.02)

        # Experimental data subplot
        df_exp = pd.DataFrame(np.column_stack((self.phi, self.counts)),
                              columns=['phi', 'I'])

        fig = px.scatter(df_exp, x='phi', y='I',
                         labels={'phi': '𝜙 (deg)', 'I': 'Counts'})
        fig.data[0].update(mode='markers+lines')
        fig.update_traces(marker=dict(color='black', size=5),
                          line=dict(color='black', width=2))

        fig_traces = []
        for trace in range(len(fig['data'])):
            fig_traces.append(fig['data'][trace])
        for traces in fig_traces:
            subfigure.append_trace(traces, row=1, col=1)

        subfigure.update_yaxes(title_text='Counts',
                               range=[0, 1.1 * self.counts.max()],
                               row=1, col=1)
        subfigure.update_xaxes(range=[self.phi.min(), self.phi.max()],
                               row=1, col=1)
        subfigure.update_layout(legend_title_text='', template='plotly_white')

        # BC lines subplot
        fig = px.line()
        for i in range(1, len(FHFGH)):
            hover_in = '<i>𝜙</i>: %{x:.3f} <br><b>F</b>:' + str(FHFGH[i - 1])
            hover_in += '<br><b>hkl</b>:' + df_in.columns[i]

            fig.add_scatter(x=df_in[df_in.columns[i]], y=df_in['ω'], name='',
                            hovertemplate=hover_in,
                            line=dict(color='Red',
                                      width=10 * FHFGH[i - 1] / np.max(FHFGH)))

            hover_out = '<i>𝜙</i>: %{x:.3f} <br><b>F</b>:' + str(FHFGH[i - 1])
            hover_out += '<br><b>hkl</b>:' + df_out.columns[i]

            fig.add_scatter(x=df_out[df_out.columns[i]], y=df_out['ω'], name='',
                            hovertemplate=hover_out,
                            line=dict(color='Blue',
                                      width=10 * FHFGH[i - 1] / np.max(FHFGH)))

        fig_traces = []
        for trace in range(len(fig["data"])):
            fig_traces.append(fig["data"][trace])
        for traces in fig_traces:
            subfigure.append_trace(traces, row=2, col=1)

        subfigure.add_shape(type='line', x0=self.phi.min(), y0=thG,
                            x1=self.phi.max(), y1=thG,
                            line=dict(color='black'),
                            row=2, col=1)
        subfigure.update_xaxes(title_text='𝜙 (deg)',
                               range=[self.phi.min(), self.phi.max()],
                               row=2, col=1,
                               showgrid=False)
        subfigure.update_yaxes(title_text='ω (deg)',
                               range=[thG - 0.1, thG + 0.1],
                               row=2, col=1,
                               showgrid=False)
        subfigure.update_layout(showlegend=False, template='plotly_white',
                                font=dict(size=16))

        subfigure.show()

    def peak_picker(self):

        """**Manual definition of peak list.**

        Notes:
                To select a peak, stop the mouse over the maximum and press `SPACE`.
                To unselect, stop the mouse over the maximum and press `DEL`.
                Feel free to zoom in or out.
                After finishing the selection, press `ENTER`. Then, press `ENTER` to close the figure.
        """

        plt.plot(self.phi, self.counts, c='k')
        plt.scatter(self.phi, self.counts, c='k', s=1)
        plt.xlabel('ɸ (deg)')
        plt.ylabel('counts')
        plt.xlim(self.phi.min(), self.phi.max())
        mk = plt.ginput(n=0, timeout=0, mouse_add=None,
                        mouse_pop=None, mouse_stop=None)
        plt.draw()
        while True:
            if plt.waitforbuttonpress(0):
                plt.close()
                break

        mk = np.array(mk)
        if len(mk) != 0:

            phi_mk = mk[:, 0]

            idx = []
            for j in range(len(phi_mk)):
                diff = np.abs(self.phi[self.peaks] - phi_mk[j])
                idx.append(int(diff.argmin()))

            self.peaks = idx

            plt.plot(self.phi[self.peaks], self.counts[self.peaks], "*b")
            plt.plot(self.phi, self.counts, c='k')
            plt.xlabel('ɸ (deg)')
            plt.ylabel('counts')
            plt.xlim(self.phi.min(), self.phi.max())
            plt.draw()
            while True:
                if plt.waitforbuttonpress(0):
                    plt.close()
                    break

    def peak_finder(self, minimum_intensity_p: float = 0.03):

        """**Automatic definition of the peak list.**

        Args:
            minimum_intensity_p (float): Minimum value of intensity to define a peak given as percentage of maximum
                value of spectrum. 3% by default.
        """

        yf = gaussian_filter1d(self.counts, 2)
        peaks, _ = find_peaks(yf, prominence=100,
                              height=self.counts.max() * minimum_intensity_p)

        self.peaks = peaks

    def data_cleaner(self, minimum_intensity_p: float = 0.2, points: int = 60):

        """**Update the peak list after automatic clean the data.**

        Args:
            minimum_intensity_p (float): Minimum value of intensity to define a peak given as percentage of maximum
                value of region. 20% by default.
            points (int): Max number of points defining a peak.
        """

        idx, i = [], 0
        for j in self.peaks:

            m, n = self.peak_definer(j, points)
            y = self.counts[m:n]

            height = (y.max() + (y[0] + y[-1]) / 2) * minimum_intensity_p
            pk, _ = find_peaks(gaussian_filter1d(y, 2), height=height)

            if len(pk) != 1:
                idx.append(i)
            i += 1

        self.peaks = np.delete(self.peaks, idx)

    def data_review(self):

        """**Plot the experimental data highlighting peaks. Delete the peaks selected by external user**.

        Notes:
                To select a peak, stop the mouse over the maximum and press `SPACE`.
                To unselect, stop the mouse over the maximum and press `DEL`.
                Feel free to zoom in or out.
                After finishing the selection, press `ENTER`. Then, press `ENTER` to close the figure.
        """

        self.simple_plot()
        mk = plt.ginput(n=0, timeout=0, mouse_add=None,
                        mouse_pop=None, mouse_stop=None)
        plt.draw()
        while True:
            if plt.waitforbuttonpress(0):
                plt.close()
                break

        mk = np.array(mk)
        if len(mk) != 0:

            phi_mk = mk[:, 0]

            idx = []
            for j in range(len(phi_mk)):
                diff = np.abs(self.phi[self.peaks] - phi_mk[j])
                idx.append(int(diff.argmin()))

            self.peaks = np.delete(self.peaks, idx)
            if self.region is not None:
                self.region = np.delete(self.region, idx, axis=0)

    def region_of_fit(self, interval: int = 6, points: int = 60):

        """**Defines the region of fit.**

        Args:
            interval (int): Number of standard deviations defining the region of fit. By default, 6-sigma is used.
            points (int): Max number of points defining a peak.
        """

        region, i = np.zeros((len(self.peaks), 4)), 0
        for j in self.peaks:
            m, n = self.peak_definer(j, points)
            x = self.phi[m:n]
            y = self.counts[m:n]

            s, d, f = fwhm(x, np.log10(y))

            N = np.abs(x - (d - interval * s)).argmin()
            M = np.abs(x - (d + interval * s)).argmin()

            region[i, :] = x[N], x[M], f, d
            i += 1

            self.region = region

    def MD_indexer(self, fname: str, M: list, Fmin: float):

        """**Finds the secondary reflection related to each case in the peak list.**

        Args:
            fname (str): Structure file name (the same used by Crystal class).
            M (list): Reference direction (eg, [1, 0, 0]).
            Fmin (float): Minimum cutoff value for W (absolute).

        Notes:
            Case two or more lattice planes are excited at the same angle with the same geometry:
            all will be displayed in the output file. If the geometries are opposite, the peak is
            not indexable (so it's excluded from the peak list). It's also not indexable the case
            with two excited reflections that distance themselves by less than 0.05º, whose smallest
            W value is at least half of the largest value. Except for these cases, the peak is indexed
            by the stronger reflection that is up to 0.1º of distance from the peak position.
        """

        M = np.array(M)

        fthg = "THG_G_" + "".join((self.primary.astype(int)).astype(str))
        fthg += "_M_" + "".join((M.astype(int)).astype(str))
        fthg += "_E_" + str(self.energy) + ".dat"

        try:
            self.BC = np.loadtxt(fthg)
        except (OSError, IOError, ValueError):
            struc = Crystal(fname)
            struc.klines(self.energy, self.primary, M, Fmin)
            self.BC = np.loadtxt(fthg)

        hkl = (self.BC[:, 0:3].astype(int)).astype(str)
        for i in range(len(hkl)):
            hkl[i] = " ".join(hkl[i])
        hkl = hkl[:, 0]

        idx = []
        s = []
        HKL = []
        for i in range(len(self.peaks)):

            ph = np.round(self.phi[self.peaks[i]], 3)
            if ph < 0:
                ph += 360

            phi_in, fin, hkl_in = self.index_find(ph, 4, hkl)
            phi_out, fout, hkl_out = self.index_find(ph, 3, hkl)

            if fin == 0 and fout == 0:
                HKL.append('')
                s.append(0)
                idx.append(i)
            elif np.abs(phi_in - phi_out) < 0.05:
                f = np.array([fin, fout])
                miller = np.array([hkl_in, hkl_out])
                diff = np.array([1, -1])
                ratio = f.min() / f.max()
                if not ratio >= 0.5:
                    j = f.argmax()
                    HKL.append(miller[j])
                    s.append(diff[j])
                else:
                    HKL.append('')
                    s.append(0)
                    idx.append(i)
            elif fin > fout:
                HKL.append(hkl_in)
                s.append(1)
            elif fout > fin:
                HKL.append(hkl_out)
                s.append(-1)

        self.index = np.column_stack((s, HKL))
        self.plot_indexation(HKL, s, idx)

        self.peaks, self.region, self.index = delete_multiple(self.peaks, self.region, self.index, idx=idx)

    def plot_indexation(self, hkl: list, s: list, idx: list):

        """**Plot the peaks and corresponding secondary reflection indexes.**

        Args:
            hkl (list): Secondary reflections ('' are the unindexed cases).
            s (list): Diffraction geometry (in-out or out-in).
            idx (list): Unindexed cases index in peak list.
        """

        self.simple_plot()
        for i in range(len(s)):

            N = np.abs(self.phi - self.region[i, 0]).argmin()
            M = np.abs(self.phi - self.region[i, 1]).argmin()

            x = self.phi[N:M]
            y = self.counts[N:M]

            if i in idx:
                plt.plot(x, y, "r")
            else:
                if s[i] == 1:
                    c = "r"
                else:
                    c = "b"
                plt.annotate(hkl[i], (self.phi[self.peaks[i]], self.counts[self.peaks[i]]), color=c)

        plt.draw()
        while True:
            if plt.waitforbuttonpress(0):
                plt.close()
                break

    def plot_peaks(self):

        """**Plot the selected peaks and corresponding region of fit.**"""

        self.simple_plot()
        for i in range(len(self.region[:, -1])):
            N = np.abs(self.phi - self.region[i, 0]).argmin()
            M = np.abs(self.phi - self.region[i, 1]).argmin()

            x = self.phi[N:M]
            y = self.counts[N:M]

            plt.plot(x, y, "r")
            plt.plot(self.phi[self.peaks[i]], y.max(), "*b")

        mk = plt.ginput(n=0, timeout=0, mouse_add=None,
                        mouse_pop=None, mouse_stop=None)

        mk = np.array(mk)
        if len(mk) != 0:

            phi_mk = mk[:, 0]

            idx = []
            for j in range(len(phi_mk)):
                diff = np.abs(self.phi[self.peaks] - phi_mk[j])
                idx.append(int(diff.argmin()))

            self.peaks = np.delete(self.peaks, idx)
            self.region = np.delete(self.region, idx, axis=0)

        plt.draw()
        while True:
            if plt.waitforbuttonpress(0):
                plt.close()
                break

    def data_fitter(self, nsamples: int = 1000):

        """**Determine the slope and slope error list.**

        Args:
            nsamples (int): Number of samples to calculate the slope error.
        """

        slope_err = np.zeros((len(self.peaks), 2))
        for i in tqdm(range(len(self.peaks))):
            N = np.abs(self.phi - self.region[i, 0]).argmin()
            M = np.abs(self.phi - self.region[i, 1]).argmin()

            x = self.phi[N:M]
            y = np.log10(self.counts[N:M])

            slope_err[i, 0] = peak_fit(x, y)

            if nsamples != 0:
                sample = pd.DataFrame(np.column_stack((x, y)))
                slope_err[i, 1] = bootstrap_resampling(sample, nsamples)
            else:
                slope_err[i, 1] = 0

        self.slope_data = slope_err

    def asymmetry_assigner(self, sbar: float = 0.1, tau: float = 0.4):

        """**Constructs the asymmetry type list.**

        Args:
            sbar (float): Minimum value of relative slope. Default = 10%.
            tau (float): Maximum value for ratio slope_error/slope. Default = 40%.
        """

        ratio = self.slope_data[:, 0] / np.abs(self.slope_data[:, 0].max())
        idx1 = np.where(np.abs(ratio) < sbar)

        if np.array(idx1).size == 0:
            pass
        else:
            self.index, self.slope_data, self.region = delete_multiple(self.index,
                                                                       self.slope_data,
                                                                       self.region,
                                                                       idx=idx1)

        ratio = self.slope_data[:, 1] / self.slope_data[:, 0]
        idx2 = np.where(np.abs(ratio) > tau)
        if np.array(idx2).size == 0:
            pass
        else:
            self.index, self.slope_data, self.region = delete_multiple(self.index,
                                                                       self.slope_data,
                                                                       self.region,
                                                                       idx=idx2)

        asy = np.zeros((len(self.slope_data[:, 0]), 2), dtype=str)
        asy[np.where(self.slope_data[:, 0] < 0), :] = 'H', 'L'
        asy[np.where(self.slope_data[:, 0] > 0), :] = 'L', 'H'
        asy = [''.join(row) for row in asy]

        self.asy = asy

    def save_output(self, fout: str = ''):

        """**Save the experimental data analysis.**

        Args:
            fout (str): File name to save. Default:`NAME_E_value_G_indexes.ext`, `NAME_E_value_G_indexes.red`

        Notes:
            .ext file presents slope, slope error and statistical properties.
            .red file presents hkl, asymmetry type and diffraction geometry. The 1st line presents the primary indexes.
        """

        sbar = self.slope_data[:, 0] / np.abs(self.slope_data[:, 0].max())
        tau = np.abs(self.slope_data[:, 1] / self.slope_data[:, 0])

        hkl = []
        for i in range(len(self.index[:, 1:])):
            hkl.append(self.index[i, -1].split('/ ')[0])

        M = np.column_stack((np.round(self.phi[self.peaks], 4),
                             self.asy, np.round(self.slope_data, 4),
                             np.round(sbar, 2), np.round(tau, 2),
                             self.index[:, 0], hkl, self.index[:, 1:]))

        H = np.zeros((len(hkl), 3))
        for i in range(len(hkl)):
            H[i, :] = np.rint(np.int64((hkl[i].split(' '))))

        N = np.column_stack((H, self.asy, self.index[:, 0]))

        header = '    phi    asy     slope   slope error   sbar (%)    tau (%)   s       hkl'

        if fout == '':
            fout = self.name.upper() + '_E' + str(self.energy)
            fout += '_G' + "".join((self.primary.astype(int)).astype(str))

        with open(fout + '.ext', 'wb') as f:
            np.savetxt(f, [], header=header)
            for line in np.matrix(M):
                np.savetxt(f, line, fmt='%10s %5s %10s %10s %10s %10s %5s %10s %32s')

        header = '      h          k         l           asy         s '
        with open(fout + '.red', 'wb') as f:
            np.savetxt(f, [], header=header)
            np.savetxt(f, np.column_stack((self.primary[0], self.primary[1], self.primary[2], self.energy)),
                       fmt='%10.1f')
            for line in np.matrix(N):
                np.savetxt(f, line, fmt='%10s %10s %10s %10s %10s')

        return fout

    def peak_definer(self, peak_number: int, points: int) -> tuple:

        """**Find the indexes of first and last points of peak.**

        Args:
            peak_number (int): Index of peak point on angles array.
            points (int): Max number of points defining a peak.

        Returns:
            tuple: First and last points of peak.
        """

        p = int(points / 2)

        if peak_number - p < 0:
            m = 0
        else:
            m = peak_number - p
        if peak_number + p > len(self.counts):
            n = len(self.counts) - 1
        else:
            n = peak_number + p

        return m, n

    def index_find(self, ph: float, col: int, hkl: np.ndarray) -> tuple:

        """**Find the closer multiple-diffraction case of specified azimuth angle.**

        Args:
            ph (float): Angle on Renninger scan.
            col (int): 3 or 4 (in-out or out-in BC lines).
            hkl (np.ndarray): Miller indices.

        Returns:
            tuple: Angle position, W amplitude, reflection indices.
        """

        k = np.where(np.abs(np.round(self.BC[:, col], 3) - ph) < 0.1)

        if not np.array(k).size == 0:
            m = np.argsort(np.abs(np.round(self.BC[k, col], 3) - ph))[::-1]
            phi_values = np.round(self.BC[k[0], col], 3)
            w_values = self.BC[k[0], -1]

            equal = np.all(phi_values == phi_values[m[0]][0])
            if equal:  # all BCs crossing in the same phi
                PHI = phi_values[m[0]][0]
                W = w_values[m[0]][0]
                miller = '/ '.join(hkl[k][m][0])

            elif phi_values[m[0]][0] - phi_values[m[0]][1] < 0.01:  # does not cross in the same phi, but close of that
                phi1 = phi_values[m[0]][0]
                phi2 = phi_values[m[0]][1]
                W1 = w_values[m[0]][0]
                W2 = w_values[m[0]][1]

                if phi1 == phi2 and W1 == W2:
                    miller = '/ '.join(hkl[k][m][0][:2])
                    PHI = phi1
                    W = W1
                elif np.abs(phi1 - phi2) < 0.05:
                    Wf = np.array([W1, W2])
                    if not Wf.min() / Wf.max() >= 0.5:
                        miller = hkl[k][m][0][Wf.argmax()]
                        PHI = phi_values[m[0]][Wf.argmax()]
                        W = w_values[m[0]][Wf.argmax()]
                    else:
                        miller = ''
                        PHI = 0
                        W = 0
                else:
                    miller = hkl[k][w_values.argmax()]
                    PHI = phi_values[w_values.argmax()]
                    W = w_values[w_values.argmax()]

            else:  # does not cross so close
                miller = hkl[k][w_values.argmax()]
                PHI = phi_values[w_values.argmax()]
                W = w_values[w_values.argmax()]
        else:  # does not cross
            miller = ''
            PHI = 0
            W = 0

        return PHI, W, miller

    def simple_plot(self) -> matplotlib.figure.Figure:

        """**Plot the experimental data highlighting peaks.**

        Returns:
            matplotlib.figure.Figure: Interactive plot.
        """

        fig = plt.figure()
        plt.plot(self.phi, self.counts, c='k')
        plt.plot(self.phi[self.peaks], self.counts[self.peaks], "*b")
        plt.xlabel('ɸ (deg)')
        plt.ylabel('counts')
        plt.xlim(self.phi.min(), self.phi.max())
        return fig
