#!/usr/bin/python3

import re

import time
from difflib import SequenceMatcher

import nglview as nv
import numpy as np
from ase import Atoms

__author__ = "Rafaela Felix"
__credits__ = {"Rafaela Felix", "Sergio Morelhão"}
__version__ = "1.0"
__maintainer__ = "Rafaela Felix"
__email__ = "rafaelafelixp@usp.br"
__status__ = "Production"


class Structure:

    """**Modeling and manipulating structural models from .in files**"""

    def __init__(self, fname: str):

        """Given an input file, returns a new `struc` object.

        Args:
            fname (str): Structure file name.
        """

        self.name = fname
        self.atoms = np.loadtxt(fname, usecols=0, dtype='<U32', skiprows=1)
        self.positions = np.loadtxt(fname, usecols=(1, 2, 3), skiprows=1)
        self.occupancy = np.loadtxt(fname, usecols=4, skiprows=1)
        self.bfactors = np.loadtxt(fname, usecols=5, skiprows=1)
        self.lattice = np.loadtxt(fname, max_rows=1)

    def visualizer_in(self) -> nv.widget.NGLWidget:

        """**Visualizes .in structure.**

        Returns:
            nv.widget.NGLWidget: System interactive visualization.
        """

        at = np.copy(self.atoms)  # just atoms (ions don't contribute to resonance)
        idx = np.where(np.char.isalpha(self.atoms) == False)
        if not np.shape(idx):
            pass
        else:
            ion = np.copy(self.atoms[idx])
            regex = re.compile('[^a-zA-Z]')

            for i in range(len(ion)):
                at[idx[0][i]] = regex.sub('', ion[i])

        system = Atoms(positions=self.positions * self.lattice[:3], symbols=at, cell=self.lattice)

        view = nv.show_ase(system)
        view.add_representation('point', selection='all', radius='0.6')
        view.add_label(color='black', scale=0.8, labelType='text',
                       labelText=[str(i) for i in range(len(system))],
                       zOffset=2.0, attachment='middle_center')
        view.add_unitcell()
        return view

    def replace_ion(self, old: str, new: str):

        """**Replace a specific atom or indexes of structure object with another specified value.**

        Args:
            old (str): Indices or element symbol to be replaced.
            new (str): Element or value to replace.

        Notes:
            There are 4 ways to pass indices as arguments: only a number, numbers separated by commas, an interval of
            indices (first and last indices separated by ":" ), and intervals separated by commas. For example, the
            following inputs are allowed: '1', '1, 2, 3', '1:3', '1:3, 5:6, 7:8'.
        """

        el = np.loadtxt('f0_CromerMann.txt', dtype='str', usecols=0)
        rank = np.array([SequenceMatcher(None, new, el[i]).ratio() for i in range(len(el))])

        if rank.max() != 1:
            print(new, 'is not included in Cromermann factors. Replaced by', el[rank.argmax()])  # future log warning
            new = el[rank.argmax()]

        if old.count(':') == 0:
            if len(old.split(',')) > 1:
                idx = [int(i) for i in old.split(',')]
                self.atoms[idx] = new
            else:
                if old.isnumeric():
                    self.atoms[int(old)] = new
                else:
                    idx = np.where(self.atoms == old)[0]
                    self.atoms[idx] = np.char.replace(self.atoms[idx], old, new)
        else:
            index = old.split(',')
            idx = [(int(i.split(':')[0]), int(i.split(':')[1]) + 1) for i in index]
            for i in range(len(idx)):
                m, n = idx[i]
                self.atoms[m:n] = new

    def replace_occupancy(self, index: str, value: str):

        """**Replace a specific set of occupancy numbers with another specified value.**

        Args:
            index (str): Indices of occupancy numbers to be replaced.
            value (str): New values of occupancy numbers.

        Notes:
            There are 4 ways to pass indices as arguments: only a number, numbers separated by commas, an interval of
            indices (first and last indices separated by ":" ), and intervals separated by commas. For example, the
            following inputs are allowed: '1', '1, 2, 3', '1:3', '1:3, 5:6, 7:8'.
        """

        if index.count(':') == 0:
            if len(index.split(',')) > 1:
                idx = [int(i) for i in index.split(',')]
                self.occupancy[idx] = value
            else:
                self.occupancy[int(index)] = value
        else:
            ind = index.split(',')
            idx = [(int(i.split(':')[0]), int(i.split(':')[1]) + 1) for i in ind]
            for i in range(len(idx)):
                m, n = idx[i]
                self.occupancy[m:n] = value

    def replace_bfactor(self, index: str, value: str):

        """**Replace a specific set of B-factors with another specified value.**

        Args:
            index (str): Indices of B-factors to be replaced.
            value (str): New values.

        Notes:
            There are 4 ways to pass indices as arguments: only a number, numbers separated by commas, an interval of
            indices (first and last indices separated by ":" ), and intervals separated by commas. For example, the
            following inputs are allowed: '1', '1, 2, 3', '1:3', '1:3, 5:6, 7:8'.
        """

        if index == ':':
            self.bfactors[:] = value
        elif index.count(':') == 0:
            if len(index.split(',')) > 1:
                idx = [int(i) for i in index.split(',')]
                self.bfactors[idx] = value
            else:
                self.bfactors[int(index)] = value
        else:
            ind = index.split(',')
            idx = [(int(i.split(':')[0]), int(i.split(':')[1]) + 1) for i in ind]
            for i in range(len(idx)):
                m, n = idx[i]
                self.bfactors[m:n] = value

    def append_atom(self, info: list):

        """**Append new atom to structure.**

        Args:
            info (list): Atom or ion symbol, fractional coordinates (x, y and z), occupancy number and B-factor.
        """

        el = np.loadtxt('f0_CromerMann.txt', dtype='str', usecols=0)
        rank = np.array([SequenceMatcher(None, info[0], el[i]).ratio() for i in range(len(el))])

        if rank.max() != 1:
            print(info[0], 'is not included in Cromermann factors. Replaced by', el[rank.argmax()])  # log warning
            info[0] = el[rank.argmax()]

        self.atoms = np.append(self.atoms, info[0])
        self.positions = np.row_stack((self.positions, info[1:4]))
        self.occupancy = np.append(self.occupancy, info[4])
        self.bfactors = np.append(self.bfactors, info[5])

    def delete_atom(self, index: str):

        """**Deletes a specific set of atoms.**

        Args:
            index (str): Indices of atoms to be deleted.

        Notes:
            There are 4 ways to pass indices as arguments: only a number, numbers separated by commas, an interval of
            indices (first and last indices separated by ":" ), and intervals separated by commas. For example, the
            following inputs are allowed: '1', '1, 2, 3', '1:3', '1:3, 5:6, 7:8'.
        """

        if index.count(':') == 0:
            if len(index.split(',')) > 1:
                idx = [int(i) for i in index.split(',')]
            else:
                idx = int(index)
            self.atoms = np.delete(self.atoms, idx, axis=0)
            self.positions = np.delete(self.positions, idx, axis=0)
            self.occupancy = np.delete(self.occupancy, idx, axis=0)
            self.bfactors = np.delete(self.bfactors, idx, axis=0)
        else:
            ind = index.split(',')
            idx = [(int(i.split(':')[0]), int(i.split(':')[1]) + 1) for i in ind]
            for i in range(len(idx)):
                m, n = idx[i]
                ind = np.arange(m, n)
                self.atoms = np.delete(self.atoms, ind, axis=0)
                self.positions = np.delete(self.positions, ind, axis=0)
                self.occupancy = np.delete(self.occupancy, ind, axis=0)
                self.bfactors = np.delete(self.bfactors, ind, axis=0)

    def save_infile(self, fout=''):

        """**Save .in file.**

        Args:
            fout (str): File name. Default:`datetime_fname`
        """

        fst_row = f"{self.lattice[0]:10.4f} {self.lattice[1]:10.4f} {self.lattice[2]:10.4f} " \
                  f"{self.lattice[3]:10.4f} {self.lattice[4]:10.4f} {self.lattice[5]:10.4f}\n"

        if fout == '':
            fout = time.strftime("%Y%m-%H_%M_%S") + '_' + self.name

        with open(fout, "w") as f:
            f.write(fst_row)
            for i in range(len(self.atoms)):
                text = f"{self.atoms[i]:>10s} {self.positions[i][0]:10.4f} " \
                       f"{self.positions[i][1]:10.4f} {self.positions[i][2]:10.4f} " \
                       f"{self.occupancy[i]:10.4f} {self.bfactors[i]:10.4f}\n"
                f.write(text)
